/**
*
* \brief CAN bootloader for SAM C21
*
* Copyright (c) 2018 Microchip Technology Inc.
*
* SPDX-License-Identifier: Apache-2.0
*
* Licensed under the Apache License, Version 2.0 (the "License"); you may
* not use this file except in compliance with the License.
* You may obtain a copy of the Licence at
* 
 * http://www.apache.org/licenses/LICENSE-2.0
* 
 * Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an AS IS BASIS, WITHOUT
* WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*
*/

//-----------------------------------------------------------------------------
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdalign.h>
#include <string.h>
#include "samc21.h"

//-----------------------------------------------------------------------------
#define BL_REQ_PORT           0 // PA
#define BL_REQ_PIN            PIN_PA19

#define CAN_TX_PORT           0 // PA
#define CAN_TX_PIN            PIN_PA24G_CAN0_TX
#define CAN_TX_MUX            MUX_PA24G_CAN0_TX

#define CAN_RX_PORT           0 // PA
#define CAN_RX_PIN            PIN_PA25G_CAN0_RX
#define CAN_RX_MUX            MUX_PA25G_CAN0_RX

#define CAN_INST              CAN0
#define CAN_GCLK_ID           CAN0_GCLK_ID
#define CAN_AHBMASK           MCLK_AHBMASK_CAN0

#define CAN_FILTER_ID         0x45a

// Nominal bit rate. The time quanta is 48 MHz / (5+1) = 8MHz.
// And each bit is (1 + NTSEG1 + 1 + NTSEG2 + 1) = 16 time quanta
// which means the bit rate is 8MHz / 16 = 500 KHz.
#define NBTP_NBRP_VALUE       5  // Nominal bit Baud Rate Prescaler
#define NBTP_NSJW_VALUE       3  // Nominal bit (Re)Synchronization Jump Width
#define NBTP_NTSEG1_VALUE     10 // Nominal bit Time segment before sample point
#define NBTP_NTSEG2_VALUE     3  // Nominal bit Time segment after sample point

// Data bit rate. The time quanta is 48 MHz / (0+1) =  48MHz.
// And each bit is (1 + DTSEG1 + 1 + DTSEG2 + 1) = 16 time quanta
// which means the bit rate is 48 MHz / 16 = 3 MHz.
#define DBTP_DBRP_VALUE       0  // Data bit Baud Rate Prescaler
#define DBTP_DSJW_VALUE       3  // Data bit (Re)Synchronization Jump Width
#define DBTP_DTSEG1_VALUE     10 // Data bit Time segment before sample point
#define DBTP_DTSEG2_VALUE     3  // Data bit Time segment after sample point

#define BOOTLOADER_SIZE       2048
#define APPLICATION_START     (FLASH_ADDR + BOOTLOADER_SIZE)
#define IMAGE_SIZE_OFFSET     (APPLICATION_START + 0x1c)
#define MAX_IMAGE_SIZE        (FLASH_SIZE - BOOTLOADER_SIZE)

#define PAGE_SIZE             NVMCTRL_PAGE_SIZE
#define ERASE_BLOCK_SIZE      NVMCTRL_ROW_SIZE
#define PAGES_IN_ERASE_BLOCK  NVMCTRL_ROW_PAGES

#define DATA_FLASH_ADDR       NVMCTRL_RWW_EEPROM_ADDR
#define DATA_FLASH_SIZE       NVMCTRL_RWW_EEPROM_SIZE

#define HEADER_SIZE           4
#define OFFSET_SIZE           4
#define SIZE_SIZE             4
#define CRC_SIZE              4
#define ARB_WORD_SIZE         4
#define MAX_CMD_SIZE          60

#define WORDS(x)              ((int)((x) / sizeof(uint32_t)))

#define ADDR_MASK             0xffffff00 // 256 bytes
#define SIZE_MASK             0xffffffe0 // 64 bytes

#define HEADER_MAGIC          0xe2
#define BL_REQUEST            0x2b620bc3

enum
{
  BL_CMD_UNLOCK    = 0xa0,
  BL_CMD_DATA      = 0xa1,
  BL_CMD_VERIFY    = 0xa2,
  BL_CMD_RESET     = 0xa3,
};

enum
{
  BL_RESP_OK        = 0x50,
  BL_RESP_ERROR     = 0x51,
  BL_RESP_INVALID   = 0x52,
  BL_RESP_CRC_OK    = 0x53,
  BL_RESP_CRC_FAIL  = 0x54,
  BL_RESP_SEQ_ERROR = 0x55,
};

static const uint8_t dlc_to_size[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 12, 16, 20, 24, 32, 48, 64 };

//-----------------------------------------------------------------------------
static void process_command(uint8_t *data, int frame_size);
static void flash_write(void);

//-----------------------------------------------------------------------------
static uint32_t *ram = (uint32_t *)HSRAM_ADDR;
static alignas(4) CanMramSidfe can_rx_standard_filter;
static alignas(4) CanMramRxbe  can_rx_buffer;
static alignas(4) CanMramTxbe  can_tx_buffer;
static alignas(4) CanMramTxefe can_tx_event_fifo;
static uint32_t unlock_addr = 0;
static uint32_t unlock_size = 0;
static uint32_t flash_addr;
static uint32_t flash_size;
static alignas(4) uint8_t flash_data[PAGE_SIZE];
static int flash_ptr;
static uint8_t data_seq;

//-----------------------------------------------------------------------------
#define CONFIGURE_PMUX(dir) \
  do { \
    PORT->Group[CAN_##dir##_PORT].PINCFG[CAN_##dir##_PIN].reg |= PORT_PINCFG_PMUXEN; \
    if (CAN_##dir##_PIN & 1) \
      PORT->Group[CAN_##dir##_PORT].PMUX[CAN_##dir##_PIN >> 1].bit.PMUXO = CAN_##dir##_MUX; \
    else \
      PORT->Group[CAN_##dir##_PORT].PMUX[CAN_##dir##_PIN >> 1].bit.PMUXE = CAN_##dir##_MUX; \
  } while (0)

//-----------------------------------------------------------------------------
static void sys_init(void)
{
  PAC->WRCTRL.reg = PAC_WRCTRL_PERID(ID_DSU) | PAC_WRCTRL_KEY_CLR;
  MCLK->APBCMASK.reg |= MCLK_AHBMASK_NVMCTRL | MCLK_AHBMASK_DSU;
  MCLK->APBBMASK.reg |= MCLK_APBBMASK_NVMCTRL | MCLK_APBBMASK_DSU;
  NVMCTRL->CTRLB.reg = NVMCTRL_CTRLB_CACHEDIS(1) | NVMCTRL_CTRLB_RWS(2) | NVMCTRL_CTRLB_MANW;
  OSCCTRL->OSC48MDIV.reg = OSCCTRL_OSC48MDIV_DIV(0);
}

//-----------------------------------------------------------------------------
static void sys_deinit(void)
{
  OSCCTRL->OSC48MDIV.reg = OSCCTRL_OSC48MDIV_DIV(12);
  NVMCTRL->CTRLB.reg = NVMCTRL_CTRLB_MANW;
  MCLK->APBCMASK.reg &= ~MCLK_AHBMASK_DSU;
  MCLK->APBBMASK.reg &= ~MCLK_APBBMASK_DSU;
  PAC->WRCTRL.reg = PAC_WRCTRL_PERID(ID_DSU) | PAC_WRCTRL_KEY_SET;
}

//-----------------------------------------------------------------------------
static uint32_t crc32_calc(uint32_t addr, uint32_t size)
{
  DSU->ADDR.reg = addr;
  DSU->LENGTH.reg = size;
  DSU->DATA.reg = 0xffffffff;
  DSU->STATUSA.reg = DSU_STATUSA_MASK;
  DSU->CTRL.reg = DSU_CTRL_CRC;

  while (0 == DSU->STATUSA.bit.DONE);

  if (DSU->STATUSA.bit.BERR)
    return 0xffffffff;

  return DSU->DATA.reg;
}

//-----------------------------------------------------------------------------
static void run_application(void)
{
  uint32_t msp = *(uint32_t *)(APPLICATION_START);
  uint32_t reset_vector = *(uint32_t *)(APPLICATION_START + 4);
  uint32_t size = *(uint32_t *)IMAGE_SIZE_OFFSET;

  if (0xffffffff == msp || size > MAX_IMAGE_SIZE)
    return;

  if (size > 0 && crc32_calc(APPLICATION_START, size) > 0)
    return;

  sys_deinit();

  __set_MSP(msp);

  asm("bx %0"::"r" (reset_vector));
}

//-----------------------------------------------------------------------------
static void delay_cycles(uint32_t cycles)
{
  cycles /= 4;

  asm volatile (
    "1: sub %[cycles], %[cycles], #1 \n"
    "   nop \n"
    "   bne 1b \n"
    : [cycles] "+l"(cycles)
  );
}

//-----------------------------------------------------------------------------
static bool bl_request(void)
{
  PORT->Group[BL_REQ_PORT].DIRCLR.reg = (1 << BL_REQ_PIN);
  PORT->Group[BL_REQ_PORT].OUTSET.reg = (1 << BL_REQ_PIN);
  PORT->Group[BL_REQ_PORT].PINCFG[BL_REQ_PIN].reg = PORT_PINCFG_INEN | PORT_PINCFG_PULLEN;

  delay_cycles(1 * (F_CPU / 1000)); // 1 ms, CPU frequency is 48 MHz at this point

  if (0 == (PORT->Group[BL_REQ_PORT].IN.reg & (1 << BL_REQ_PIN)))
    return true;

  if (BL_REQUEST == ram[0] && BL_REQUEST == ram[1] &&
      BL_REQUEST == ram[2] && BL_REQUEST == ram[3])
  {
    ram[0] = 0;
    return true;
  }

  return false;
}

//-----------------------------------------------------------------------------
static void can_init(void)
{
  CONFIGURE_PMUX(RX);
  CONFIGURE_PMUX(TX);

  MCLK->AHBMASK.reg |= CAN_AHBMASK;

  GCLK->PCHCTRL[CAN_GCLK_ID].reg = GCLK_PCHCTRL_GEN(0) | GCLK_PCHCTRL_CHEN;
  while (0 == (GCLK->PCHCTRL[CAN_GCLK_ID].reg & GCLK_PCHCTRL_CHEN));

  CAN_INST->CCCR.reg   = CAN_CCCR_INIT | CAN_CCCR_CCE;
  CAN_INST->CCCR.reg   = CAN_CCCR_INIT | CAN_CCCR_CCE | CAN_CCCR_TXP |
      CAN_CCCR_EFBI | CAN_CCCR_PXHD | CAN_CCCR_DAR | CAN_CCCR_FDOE | CAN_CCCR_BRSE;
  CAN_INST->SIDFC.reg  = CAN_SIDFC_FLSSA((uint32_t)&can_rx_standard_filter) | CAN_SIDFC_LSS(1);
  CAN_INST->XIDFC.reg  = 0;
  CAN_INST->RXF0C.reg  = CAN_RXF0C_F0OM | CAN_RXF0C_F0WM(0);
  CAN_INST->RXF1C.reg  = CAN_RXF1C_F1OM | CAN_RXF1C_F1WM(0);
  CAN_INST->RXBC.reg   = CAN_RXBC_RBSA((uint32_t)&can_rx_buffer);
  CAN_INST->TXBC.reg   = CAN_TXBC_TBSA((uint32_t)&can_tx_buffer) | CAN_TXBC_NDTB(1);
  CAN_INST->TXEFC.reg  = CAN_TXEFC_EFSA((uint32_t)&can_tx_event_fifo) | CAN_TXEFC_EFS(1);
  CAN_INST->RXESC.reg  = CAN_RXESC_RBDS_DATA64 | CAN_RXESC_F0DS_DATA64 | CAN_RXESC_F1DS_DATA64;
  CAN_INST->TXESC.reg  = CAN_TXESC_TBDS_DATA64;
  CAN_INST->TSCC.reg   = CAN_TSCC_TCP(0) | CAN_TSCC_TSS_INC_Val;
  CAN_INST->TDCR.reg   = CAN_TDCR_TDCO(0) | CAN_TDCR_TDCF(0);
  CAN_INST->GFC.reg    = CAN_GFC_ANFS_REJECT | CAN_GFC_ANFE_REJECT | CAN_GFC_RRFS | CAN_GFC_RRFE;
  CAN_INST->XIDAM.reg  = CAN_XIDAM_MASK;
  CAN_INST->TXBTIE.reg = CAN_TXBTIE_MASK;
  CAN_INST->TXBCIE.reg = CAN_TXBCIE_MASK;

  CAN_INST->NBTP.reg = CAN_NBTP_NBRP(NBTP_NBRP_VALUE) |
      CAN_NBTP_NSJW(NBTP_NSJW_VALUE) |
      CAN_NBTP_NTSEG1(NBTP_NTSEG1_VALUE) |
      CAN_NBTP_NTSEG2(NBTP_NTSEG2_VALUE);
  CAN_INST->DBTP.reg = CAN_DBTP_DBRP(DBTP_DBRP_VALUE) |
      CAN_DBTP_DSJW(DBTP_DSJW_VALUE) |
      CAN_DBTP_DTSEG1(DBTP_DTSEG1_VALUE) |
      CAN_DBTP_DTSEG2(DBTP_DTSEG2_VALUE);

  CAN_INST->CCCR.reg &= ~CAN_CCCR_INIT;
  while (CAN_INST->CCCR.reg & CAN_CCCR_INIT);

  can_rx_standard_filter.SIDFE_0.reg = CAN_SIDFE_0_SFID2(0) | CAN_SIDFE_0_SFID1(CAN_FILTER_ID) |
      CAN_SIDFE_0_SFT_CLASSIC | CAN_SIDFE_0_SFEC_STRXBUF;
}

//-----------------------------------------------------------------------------
static void send_response(int code)
{
  uint8_t *data = (uint8_t *)can_tx_buffer.TXBE_DATA;

  data[0] = code;

  can_tx_buffer.TXBE_0.reg = CAN_TXBE_0_ID(CAN_FILTER_ID << 18);
  can_tx_buffer.TXBE_1.reg = CAN_TXBE_1_FDF | CAN_TXBE_1_BRS | CAN_TXBE_1_DLC(1);

  CAN_INST->TXBAR.reg = CAN_TXBAR_AR0;

  #define TX_END_FLAGS (CAN_IR_TC | CAN_IR_TOO | CAN_IR_PEA | CAN_IR_PED)

  while (0 == (CAN_INST->IR.reg & TX_END_FLAGS));

  CAN_INST->IR.reg = TX_END_FLAGS;
}

//-----------------------------------------------------------------------------
static void can_task(void)
{
  uint8_t *data = (uint8_t *)can_rx_buffer.RXBE_DATA;
  int size;

  if (0 == CAN_INST->IR.bit.DRX)
    return;

  CAN_INST->IR.reg = CAN_IR_DRX;

  if (0 == CAN_INST->NDAT1.bit.ND0)
    return;

  size = dlc_to_size[can_rx_buffer.RXBE_1.bit.DLC];

  process_command(data, size);

  CAN_INST->NDAT1.reg = CAN_NDAT1_ND0;
}

//-----------------------------------------------------------------------------
static void process_command(uint8_t *data, int frame_size)
{
  uint32_t *wdata = (uint32_t *)data;
  int command, size;

  // Ignore frames that don't have a complete header, since we
  // can't consistently respond to them.
  if (frame_size < HEADER_SIZE)
    return;

  command = data[0];
  size = data[3];

  if ((size > MAX_CMD_SIZE) || (frame_size < (HEADER_SIZE + size)) || (HEADER_MAGIC != data[2]))
    return send_response(BL_RESP_ERROR);

  if (BL_CMD_UNLOCK == command)
  {
    uint32_t uaddr = wdata[1] & ADDR_MASK;
    uint32_t usize = wdata[2] & SIZE_MASK;
    bool pass = (usize > 0) && ((usize <= FLASH_SIZE) ||
        (uaddr >= DATA_FLASH_ADDR && usize <= DATA_FLASH_SIZE));

    if (size != (OFFSET_SIZE + SIZE_SIZE))
      return send_response(BL_RESP_ERROR);

    if (pass)
    {
      unlock_addr = uaddr;
      unlock_size = usize;
      send_response(BL_RESP_OK);
    }
    else
    {
      unlock_addr = 0;
      unlock_size = 0;
      send_response(BL_RESP_ERROR);
    }

    data_seq = 0;
    flash_ptr = 0;
    flash_addr = unlock_addr;
    flash_size = unlock_size;
  }
  else if (BL_CMD_DATA == command)
  {
    if (data[1] != data_seq)
      return send_response(BL_RESP_SEQ_ERROR);

    for (int i = 0; i < size; i++)
    {
      if (0 == flash_size)
        return send_response(BL_RESP_ERROR);

      flash_data[flash_ptr++] = data[4 + i];

      if (flash_ptr == PAGE_SIZE)
      {
        flash_write();

        flash_ptr = 0;
        flash_addr += PAGE_SIZE;
        flash_size -= PAGE_SIZE;
      }
    }

    data_seq++;
    send_response(BL_RESP_OK);
  }
  else if (BL_CMD_VERIFY == command)
  {
    uint32_t crc;

    if (size != CRC_SIZE)
      return send_response(BL_RESP_ERROR);

    crc = crc32_calc(unlock_addr, unlock_size);

    if (wdata[1] == crc)
      send_response(BL_RESP_CRC_OK);
    else
      send_response(BL_RESP_CRC_FAIL);
  }
  else if (BL_CMD_RESET == command)
  {
    if (size != (ARB_WORD_SIZE * 4))
      return send_response(BL_RESP_ERROR);

    ram[0] = wdata[1];
    ram[1] = wdata[2];
    ram[2] = wdata[3];
    ram[3] = wdata[4];

    send_response(BL_RESP_OK);

    NVIC_SystemReset();
  }
  else
  {
    send_response(BL_RESP_INVALID);
  }
}

//-----------------------------------------------------------------------------
static void flash_write(void)
{
  uint32_t *src = (uint32_t *)flash_data;
  uint32_t *dst = (uint32_t *)flash_addr;

  NVMCTRL->ADDR.reg = flash_addr >> 1;

  if (0 == (flash_addr % ERASE_BLOCK_SIZE))
  {
    // Lock region size is always bigger than the row size
    NVMCTRL->CTRLA.reg = NVMCTRL_CTRLA_CMDEX_KEY | NVMCTRL_CTRLA_CMD_UR;
    while (0 == NVMCTRL->INTFLAG.bit.READY);

    NVMCTRL->CTRLA.reg = NVMCTRL_CTRLA_CMDEX_KEY | NVMCTRL_CTRLA_CMD_ER;
    while (0 == NVMCTRL->INTFLAG.bit.READY);
  }

  for (int i = 0; i < WORDS(PAGE_SIZE); i++) 
    dst[i] = src[i];

  NVMCTRL->CTRLA.reg = NVMCTRL_CTRLA_CMDEX_KEY | NVMCTRL_CTRLA_CMD_WP;
  while (0 == NVMCTRL->INTFLAG.bit.READY);
}

//-----------------------------------------------------------------------------
__attribute__ ((noinline)) // Prevent LTO from inlining main() into the reset handler
int main(void)
{
  sys_init();

  if (!bl_request())
    run_application();

  can_init();

  while (1)
  {
    can_task();
  }

  return 0;
}
